package andy.hornetq;

import static org.junit.Assert.assertEquals;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicInteger;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Queue;
import javax.jms.Session;
import javax.jms.Topic;

import org.hornetq.api.core.TransportConfiguration;
import org.hornetq.core.config.Configuration;
import org.hornetq.core.config.impl.ConfigurationImpl;
import org.hornetq.core.remoting.impl.netty.NettyAcceptorFactory;
import org.hornetq.core.remoting.impl.netty.NettyConnectorFactory;
import org.hornetq.jms.server.config.JMSConfiguration;
import org.hornetq.jms.server.config.impl.ConnectionFactoryConfigurationImpl;
import org.hornetq.jms.server.config.impl.JMSConfigurationImpl;
import org.hornetq.jms.server.config.impl.JMSQueueConfigurationImpl;
import org.hornetq.jms.server.config.impl.TopicConfigurationImpl;
import org.hornetq.jms.server.embedded.EmbeddedJMS;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

public class MultipleAckTest {
	private static final int MESSAGES_TO_SEND_PER_PRODUCER = 1000;
	private static final int MULTIPLE_PRODUCERS_COUNT = 2;
	// number of milliseconds to wait to receive a message...
	private static final long CONSUMER_WAIT_TIME_BEFORE_TERMINATE = 500;
	
	private static EmbeddedJMS jmsServer;
	private static ConnectionFactory connectionFactory;
	private static Topic nonDurableTopic;
	private static Topic durableTopic;
	private static Queue queue;
	
	private final AtomicInteger messagesConsumed = new AtomicInteger();
	private final AtomicInteger messagesProduced= new AtomicInteger();
	private final Map<String, AtomicInteger> allMessageIds = new HashMap<String, AtomicInteger>();
	// latch to indicate when the consumer is up and running...
	private final CountDownLatch consumerStartedCountDownLatch = new CountDownLatch(1);
	// latch to indicate when all of the threads have finished processing...
	private CountDownLatch threadsRunningCountDownLatch;

	@BeforeClass
	public static void initialiseHornetQ() throws Exception {
		final Configuration config = new ConfigurationImpl();
		config.setPersistenceEnabled(false);
		config.setSecurityEnabled(false);
		config.setAcceptorConfigurations(Collections.singleton(new TransportConfiguration(NettyAcceptorFactory.class.getName())));
		config.setConnectorConfigurations(Collections.singletonMap("netty", new TransportConfiguration(NettyConnectorFactory.class.getName())));

		final JMSConfiguration jmsConfig = new JMSConfigurationImpl();
		jmsConfig.getConnectionFactoryConfigurations().add(new ConnectionFactoryConfigurationImpl("cf", false, Arrays.asList("netty"), "/cf"));
		
		jmsConfig.getTopicConfigurations().add(new TopicConfigurationImpl("nonDurableTopic", "/topic/nonDurableTopic"));
		jmsConfig.getTopicConfigurations().add(new TopicConfigurationImpl("durableTopic", "/topic/durableTopic"));
		jmsConfig.getQueueConfigurations().add(new JMSQueueConfigurationImpl("simple", "", false, "/queue/simple"));
		
		jmsServer = new EmbeddedJMS();
		jmsServer.setConfiguration(config);
		jmsServer.setJmsConfiguration(jmsConfig);
		jmsServer.start();
		
		connectionFactory = (ConnectionFactory) jmsServer.lookup("/cf");
		nonDurableTopic = (Topic) jmsServer.lookup("/topic/nonDurableTopic");
		durableTopic = (Topic) jmsServer.lookup("/topic/durableTopic");
		queue = (Queue) jmsServer.lookup("/queue/simple");
	}
	
	@AfterClass
	public static void stopHornetQ() throws Exception {
		if (jmsServer != null) {
			jmsServer.stop();
			jmsServer = null;
			connectionFactory = null;
			nonDurableTopic = null;
			durableTopic = null;
			queue = null;
		}
	}

	@Test
	public void testNonDurableTopicMultipleProducers() throws Exception {
		runTest(nonDurableTopic, false, MULTIPLE_PRODUCERS_COUNT);
	}

	@Test
	public void testDurableTopicMultipleProducers() throws Exception {
		runTest(durableTopic, true, MULTIPLE_PRODUCERS_COUNT);
	}

	@Test
	public void testQueueMultipleProducers() throws Exception {
		runTest(queue, false, MULTIPLE_PRODUCERS_COUNT);
	}

	private void runTest(final Destination destination,
						 final boolean durable,
						 final int producersCount) throws Exception {
		this.threadsRunningCountDownLatch = new CountDownLatch(producersCount + 1);
		createConsumer(destination, durable).start();
		
		for (int i = 0; i < producersCount; i++) {
			createProducer(destination, MESSAGES_TO_SEND_PER_PRODUCER).start();
		}
		
		// wait for all of the running threads to finish before we validate...
		this.threadsRunningCountDownLatch.await();
		
		// log out any duplicate messages
		for (Map.Entry<String, AtomicInteger> entry : this.allMessageIds.entrySet()) {
			if (entry.getValue().intValue() > 1) {
				System.err.println("!!!!The message " + entry.getKey() + " was received " + entry.getValue() + " times");
			}
		}

		final int totalExpectedMEssages = producersCount * MESSAGES_TO_SEND_PER_PRODUCER;
		assertEquals(totalExpectedMEssages, this.allMessageIds.size());
		assertEquals(totalExpectedMEssages, this.messagesConsumed.intValue());
		assertEquals(totalExpectedMEssages, this.messagesProduced.intValue());
	}

	private Thread createConsumer(final Destination destination,
								  final boolean durable) {
		return new SessionThread() {
			@Override
			protected void withSession(final Session session) throws Exception {
				final MessageConsumer consumer;
				
				if (durable) {
					consumer = session.createDurableSubscriber((Topic) destination, getClass().getSimpleName() + "-" + System.identityHashCode(this)); 
				} else {
					consumer = session.createConsumer(destination);
				}

				try {
					consumerStartedCountDownLatch.countDown();
					
					while (true) {
						final Message message = consumer.receive(CONSUMER_WAIT_TIME_BEFORE_TERMINATE);
						
						// exit when no messages have been received before we timed out...
						if (message == null) {
							break;
						}
						
						messagesConsumed.incrementAndGet();
						
						final String messageId = message.getJMSMessageID();
						final AtomicInteger ai = allMessageIds.get(messageId);
						
						if (ai == null) {
							allMessageIds.put(messageId, new AtomicInteger(1));
						} else {
							ai.incrementAndGet();
						}
					}
				} finally {
					consumer.close();
				}
			}
		};
	}

	private Thread createProducer(final Destination destination,
								  final int messagesToSend) {
		return new SessionThread() {
			@Override
			protected void withSession(final Session session) throws Exception {
				// before we continue, wait for the consumer to start...
				consumerStartedCountDownLatch.await();

				final MessageProducer producer = session.createProducer(destination);
				
				try {
					for (int i = 0; i < messagesToSend; i++) {
						producer.send(session.createTextMessage(getClass().getSimpleName() + "-" + System.identityHashCode(this) + "-" + i));
						messagesProduced.incrementAndGet();
					}
				} finally {
					producer.close();
				}
			}
		};
	}

	
	private abstract class SessionThread extends Thread {
		@Override
		public void run() {
			Connection con = null;
			Session session = null;
			
			try {
				con = connectionFactory.createConnection();
				con.setClientID(getClass().getSimpleName() +"-" + System.identityHashCode(this));
				con.start();
				session = con.createSession(false, Session.AUTO_ACKNOWLEDGE);

				withSession(session);
			} catch (final Exception e) {
				e.printStackTrace();
			} finally {
				if (session != null) {
					try {
						session.close();
					} catch (final Exception e) {
						e.printStackTrace();
					}
				}

				if (con != null) {
					try {
						con.close();
					} catch (final Exception e) {
						e.printStackTrace();
					}
				}
			}
			
			threadsRunningCountDownLatch.countDown();
		}

		protected abstract void withSession(Session session) throws Exception;
	}
}
